import logging
from socketserver import ThreadingMixIn
from wsgiref.simple_server import WSGIServer, make_server
from django.core.management.base import BaseCommand
from prometheus_client import make_wsgi_app

from promy.utils.monitoring import REGISTRY

logger = logging.getLogger(__name__)


class ThreadingWSGIServer(ThreadingMixIn, WSGIServer):
    """Thread per request HTTP server."""

    # Make worker threads "fire and forget". Beginning with Python 3.7 this
    # prevents a memory leak because ``ThreadingMixIn`` starts to gather all
    # non-daemon threads in a list in order to join on them at server close.
    daemon_threads = True


class Command(BaseCommand):
    help = "Serve prometheus metrics endpoint"

    def add_arguments(self, parser):
        parser.add_argument("addr", type=str)
        parser.add_argument("port", type=int)

    def handle(self, *args, **options):
        app = make_wsgi_app(REGISTRY)
        httpd = make_server(options["addr"], options["port"], app, ThreadingWSGIServer)
        logger.info(f"Metrics are exposed on port {options['port']}...\n")
        httpd.serve_forever()



from rest_framework import serializers


class BaseModelSerializer(serializers.ModelSerializer):
    CREATED_AT = False
    UPDATED_AT = False

    created_at = serializers.DateTimeField(read_only=True)
    updated_at = serializers.DateTimeField(read_only=True)
    key = serializers.UUIDField(read_only=True)

    class Meta:
        abstract = True

    def get_fields(self):
        fields = super().get_fields()

        if "archived_at" in fields:
            fields.pop("archived_at")

        if not self.CREATED_AT and "created_at" in fields:
            fields.pop("created_at")

        if not self.UPDATED_AT and "updated_at" in fields:
            fields.pop("updated_at")

        return fields


class UserIdSerializerMixin:
    def validate(self, attrs):
        attrs["user_id"] = self.context.get("user_id", None)
        return attrs

    def add_user_id(self, user_id):
        self.context["user_id"] = user_id



from typing import Any


def getpath(o: Any, path: str, default=None) -> Any:
    for p in path.split("."):
        if o is None:
            return default
        o = getattr(o, p, None)
    return default if o is None else o



from django.utils.timezone import now
from rest_framework import status
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.viewsets import ModelViewSet

__all__ = ["BaseModelViewSet", "UserIdFilterViewMixin", "KwargsContextViewMixin"]


class UserIdFilterViewMixin:
    def get_queryset(self):
        return super().get_queryset().filter(user_id=self.kwargs.get("user_id"))


class KwargsContextViewMixin:
    def get_serializer_context(self):
        return self.kwargs


class BaseModelViewSet(KwargsContextViewMixin, ModelViewSet):
    request_serializer = None
    response_serializer = None

    def get_serializer_class(self):
        if self.action in ["create", "update", "partial_update"]:
            return self.request_serializer or self.serializer_class

        return self.response_serializer or self.serializer_class

    def destroy(self, request: Request, *args, **kwargs):
        instance = self.get_object()

        if hasattr(instance, "archived_at"):
            instance.archived_at = now()
            instance.save()
        else:
            self.perform_destroy(instance)

        return Response(status=status.HTTP_200_OK)



import json
import logging
from functools import cached_property
from time import perf_counter as time_now

from django.conf import settings
from django.core.serializers.json import DjangoJSONEncoder
from kafka.producer import KafkaProducer as KafkaProducerClient

from promy.utils.monitoring import KAFKA_PRODUCE_DURATION
from promy.utils.tracing import with_trace

logger = logging.getLogger(__name__)


class KafkaProducer:
    def __init__(self, topic_name: str):
        self.topic_name = topic_name

    @cached_property
    def instance(
        self,
    ) -> KafkaProducerClient:
        kafka = settings.ENV.kafka
        assert kafka.enabled

        return KafkaProducerClient(
            bootstrap_servers=kafka.bootstrap_servers,
            security_protocol=kafka.security_protocol,
            sasl_mechanism=kafka.sasl_mechanism,
            sasl_plain_username=kafka.username,
            sasl_plain_password=kafka.password,
            api_version=(0, 10, 2),
            value_serializer=lambda m: json.dumps(m, cls=DjangoJSONEncoder).encode(
                "utf-8"
            ),
            key_serializer=lambda k: k.encode("utf-8"),
        )

    @with_trace("produce")
    def produce(self, key: str, value: dict, should_close: bool = False):
        logger.info(
            f"Attempting to send message to kafka: topic={self.topic_name}, {key=}, {value=}"
        )

        start = time_now()
        success = "false"

        retries = 0
        while retries < 5:
            try:
                self.instance.send(self.topic_name, value, key)
                success = "true"
            except Exception as e:
                logger.exception(e)
                retries += 1
            else:
                break

        KAFKA_PRODUCE_DURATION.labels(self.topic_name, success).observe(
            time_now() - start
        )

        if should_close:
            self.instance.close()

    def close(self):
        self.instance.close()



import json
import logging

import sentry_sdk
from time import perf_counter as time_now
from django.conf import settings
from django.core.management.base import BaseCommand
from django.db import Error as DbError
from kafka import KafkaConsumer

from promy.station.registry import station_registry
from promy.utils.monitoring import KAFKA_CONSUME_DURATION

logger = logging.getLogger("Consumer")


class Command(BaseCommand):
    def handle(self, *args, **options):
        handlers = station_registry.kafka_updater_handlers

        kafka = settings.ENV.kafka

        while True:
            consumer = KafkaConsumer(
                *handlers.keys(),
                bootstrap_servers=kafka.bootstrap_servers,
                security_protocol=kafka.security_protocol,
                sasl_mechanism=kafka.sasl_mechanism,
                sasl_plain_username=kafka.username,
                sasl_plain_password=kafka.password,
                group_id=kafka.group_id,
                enable_auto_commit=False,
                auto_offset_reset="earliest",
                value_deserializer=lambda x: json.loads(x),
                key_deserializer=lambda x: x.decode(),
            )

            for message in consumer:
                sentry_sdk.set_context(
                    "message",
                    {
                        "topic": message.topic,
                        "partition": message.partition,
                        "offset": message.offset,
                        "key": message.key,
                        "value": message.value,
                    },
                )

                start = time_now()
                success = "false"

                try:
                    handlers[message.topic](message.value)
                    success = "true"
                except DbError:
                    logger.exception(
                        "handler failed due to database issues, exiting",
                        extra={
                            "topic": message.topic,
                            "partition": message.partition,
                            "offset": message.offset,
                            "key": message.key,
                        },
                    )
                    exit(1)
                except Exception:
                    logger.exception(
                        "handler failed",
                        extra={
                            "topic": message.topic,
                            "partition": message.partition,
                            "offset": message.offset,
                            "key": message.key,
                        },
                    )
                else:
                    consumer.commit()
                    logger.info(
                        "consumer committed",
                        extra={
                            "topic": message.topic,
                            "partition": message.partition,
                            "offset": message.offset,
                            "key": message.key,
                        },
                    )

                KAFKA_CONSUME_DURATION.labels(message.topic, success).observe(
                    time_now() - start
                )



from django.conf import settings
from .python import getpath


def is_testing() -> bool:
    return getattr(settings, "TESTING", False)


def get_environment() -> str | None:
    if e := getpath(settings, "ENV.environment"):
        return e
    return getpath(settings, "ENV.sentry.environment")



import logging

from django.core.management.base import BaseCommand

from promy.station.registry import station_registry
from promy.utils.periodic_tasks import set_periodic_tasks

logger = logging.getLogger("Consumer")


class Command(BaseCommand):
    def handle(self, *args, **options):
        if station_registry.periodic_tasks is not None:
            set_periodic_tasks(
                station_registry.periodic_tasks,
                station_registry.periodic_tasks_disable_rest,
            )



import logging

logger = logging.getLogger(__name__)


class StationRegistry:
    def __init__(self):
        self.publishers: list = []
        self.updater_handlers: list = []
        self.kafka_updater_handlers: dict = {}
        self.periodic_updaters = None
        self.periodic_tasks = None
        self.periodic_tasks_disable_rest = True

    def set_publishers(self, publishers: list) -> None:
        self.publishers = publishers

    def set_updater_handlers(self, handlers: list) -> None:
        self.updater_handlers = handlers

    def set_kafka_updater_handlers(
        self, handlers: dict
    ) -> None:  # topic_name -> UpdaterHandler
        self.kafka_updater_handlers = handlers

    def set_periodic_updaters(self, updaters):
        self.periodic_updaters = updaters

    def set_periodic_tasks(self, tasks: dict | list, disable_rest: bool = True) -> None:
        self.periodic_tasks = tasks
        self.periodic_tasks_disable_rest = disable_rest


station_registry = StationRegistry()



import json
import logging
from typing import Any

from django.core.management.base import BaseCommand
from django_celery_beat.models import PeriodicTask, IntervalSchedule

from promy.utils.periodic_tasks import (
    PromTasks,
    _UNSUPPORTED_PERIODIC_TASK_FIELDS,
    _ALL_PERIODIC_TASK_FIELDS,
)

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Export currently enabled PeriodicTasks"

    def handle(self, *args, **options):
        tasks = []

        for task in PeriodicTask.objects.filter(enabled=True).order_by("id"):
            if task.name == "celery.backend_cleanup":
                continue

            skip = False

            for field in _UNSUPPORTED_PERIODIC_TASK_FIELDS:
                if getattr(task, field) is not None:
                    skip = True
                    logger.warning(
                        f"Unsupported field {field} found in task {task.name}, skipping this task"
                    )
                    break
            if skip:
                continue

            task_dict = {}
            for field, default in _ALL_PERIODIC_TASK_FIELDS.items():
                value = getattr(task, field, None)
                if value == default or field in ("start_time",):
                    continue

                task_dict[field] = self._get_value(field, value)

            tasks.append(task_dict)

        result = json.dumps(tasks, indent=4)

        for promy_task, task_path in PromTasks.as_dict().items():
            result = result.replace(f'"{task_path}"', f"PromTasks.{promy_task}")

        print(result)

    def _get_value(self, key: str, value: Any) -> Any:
        if key == "interval":
            return self._serialize_interval_schedule(value)
        return value

    @staticmethod
    def _serialize_interval_schedule(interval: IntervalSchedule) -> str:
        unit_mapping = {
            "days": "d",
            "hours": "h",
            "minutes": "m",
            "seconds": "s",
            "microseconds": "us",
        }
        return f"{interval.every}{unit_mapping[interval.period]}"



import hashlib
import json
from datetime import timedelta
from typing import Callable, Optional, Union, Any

from django.conf import settings
from django.core.serializers.json import DjangoJSONEncoder
from django.db.models import QuerySet
from django.utils import timezone
from rest_framework.serializers import SerializerMetaclass

from promy.utils.redis import get_redis_client

Duration = Union[int, float, timedelta]

RECENTLY_UPDATED_THRESHOLD = timedelta(minutes=15)
CACHE_EXPIRATION = timedelta(days=1)


class Publisher:
    def __init__(
        self,
        *,
        name: str | None = None,
        queryset: QuerySet,
        serializer: SerializerMetaclass | Callable | None = None,
        publish_to_broker: Optional[Callable[[str, dict], None]] = None,
        publish: Optional[Callable] = None,
        checksum: Optional[Callable] = None,
        recently_updated_threshold: Duration = RECENTLY_UPDATED_THRESHOLD,
        cache_expiration: Duration | None = CACHE_EXPIRATION,
        enable_publish_all: bool = True,
    ):
        self._publish: Optional[Callable] = publish
        self.name: str | None = name
        self.queryset = queryset
        self.serializer: SerializerMetaclass | Callable | None = serializer
        self.publish_to_broker: Optional[Callable] = publish_to_broker
        self.checksum: Optional[Callable] = checksum
        self.recently_updated_threshold: Duration = recently_updated_threshold
        self.cache_expiration: Duration | None = cache_expiration
        self.enable_publish_all = enable_publish_all

        if not publish:
            assert name, "name is required"
            assert serializer is not None, "serializer is required"
            assert publish_to_broker, "publish_to_broker is required"

    def publish(self, instance, *args, force: bool = False, **kwargs):
        if self._publish:
            self._publish(instance, *args, **kwargs)
            return

        if getattr(settings, "TESTING", False):
            return

        assert self.name
        assert self.serializer
        assert self.publish_to_broker

        r = get_redis_client()

        object_key = str(instance.key)

        with r.lock(f"publisher_lock:{self.name}:{object_key}", timeout=10):
            hash_key = f"publisher_last_msg:{self.name}:{object_key}"
            checksum = self.get_checksum(instance)

            if not force and r.get(hash_key) == checksum:
                return

            self.publish_to_broker(object_key, self.get_instance_message(instance))
            r.set(
                hash_key,
                checksum,
                ex=(
                    int(_duration_to_seconds(self.cache_expiration))
                    if self.cache_expiration
                    else None
                ),
            )

    def get_instance_message(self, instance: Any) -> dict:
        assert self.name
        assert self.serializer

        if isinstance(self.serializer, SerializerMetaclass):
            data = self.serializer(instance).data
        else:
            data = self.serializer(instance)

        data = json.loads(
            json.dumps(data, cls=DjangoJSONEncoder)
        )  # To convert OrderedDicts/Pydantic objects to simple dicts.
        d = {
            "object_name": self.name,
            "body": data,
        }

        return d

    def get_checksum(self, instance):
        if self.checksum:
            return self.checksum(instance)
        return (
            hashlib.sha256(
                json.dumps(
                    self.get_instance_message(instance), cls=DjangoJSONEncoder
                ).encode()
            )
            .hexdigest()
            .encode()
        )

    def _publish_queryset(self, queryset: QuerySet, *args, **kwargs):
        for instance in queryset.all():
            self.publish(instance, *args, **kwargs)

    def publish_all(self, *args, **kwargs):
        if self.enable_publish_all:
            self._publish_queryset(self.queryset, *args, **kwargs)

    def publish_recently_updated(
        self, threshold: Duration | None = None, *args, **kwargs
    ):
        if threshold is None:
            threshold = self.recently_updated_threshold
        delta = _duration_to_timedelta(threshold)
        self._publish_queryset(
            self.queryset.filter(updated_at__gte=timezone.now() - delta),
            *args,
            **kwargs,
        )


def _duration_to_seconds(d: Duration) -> float:
    if isinstance(d, timedelta):
        return d.total_seconds()
    return float(d)


def _duration_to_timedelta(d: Duration) -> timedelta:
    if isinstance(d, timedelta):
        return d
    return timedelta(seconds=d)



import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from promy.utils.http_requests import HTTPClient


def create_session(
    service_name="unset",
    retry_count=1,
    backoff_factor=0.5,
    **kwargs,
) -> HTTPClient:
    """
        Reference:
            https://stackoverflow.com/a/47475019/2581953
    :param retry_count:
    :param backoff_factor:
    :param service_name:
    :return:
    """
    session = requests.Session()
    retry = Retry(total=retry_count, backoff_factor=backoff_factor)
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return HTTPClient(service_name, session, **kwargs)



import hashlib
from functools import wraps
from typing import Any, get_args, Callable

from django.apps import apps
from django.core.cache import cache
from django.db.models import Model as DjangoModel
from pydantic import BaseModel as PydanticModel


# get the cache key for storage
def cache_get_key(*args, **kwargs) -> str:
    serialise = []
    for arg in args:
        serialise.append(str(arg))
    for key, arg in kwargs.items():
        serialise.append(str(key))
        serialise.append(str(arg))
    key = hashlib.md5(":".join(serialise).encode("utf-8")).hexdigest()
    return key


# decorator for caching functions
def cache_for(seconds: int | float, cache_none: bool = True) -> Callable:
    def decorator(fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            key = cache_get_key(fn.__module__ + "." + fn.__name__, *args, **kwargs)
            if cached_result := cache.get(key):
                return _deserialize(
                    cached_result,
                    _get_all_pydantic_models(fn.__annotations__.get("return")),
                )

            result = fn(*args, **kwargs)

            if not cache_none and result is None:
                raise ValueError("Can't cache null value")

            cache.set(key, _serialize(result), seconds)

            return result

        return wrapper

    return decorator


_NONE_VALUE = {"__none__": True}


def _serialize(data: Any, depth: int = 0) -> Any:
    if data is None and depth == 0:
        return _NONE_VALUE

    if isinstance(data, PydanticModel):
        return {"__pydantic__": data.__class__.__name__, "data": data.model_dump()}

    elif isinstance(data, DjangoModel):
        app_name = data._meta.app_label  # type: ignore
        model_name = data._meta.model_name  # type: ignore
        return {
            "__django__": {"app_name": app_name, "model_name": model_name},
            "pk": data.pk,
        }

    elif isinstance(data, (list, tuple)):
        return [_serialize(item, depth + 1) for item in data]

    elif isinstance(data, dict):
        return {key: _serialize(value, depth + 1) for key, value in data.items()}

    return data


def _get_all_pydantic_models(data_type: Any) -> dict[str, PydanticModel]:
    if data_type is None:
        return {}

    if isinstance(data_type, type) and issubclass(data_type, PydanticModel):
        return {data_type.__name__: data_type}  # type: ignore

    result = {}
    for cls in get_args(data_type):
        result.update(_get_all_pydantic_models(cls))
    return result


def _deserialize(data: Any, pydantic_models: dict) -> Any:
    if data == _NONE_VALUE:
        return None

    if isinstance(data, dict) and "__pydantic__" in data:
        if cls := pydantic_models.get(data["__pydantic__"]):
            return cls(**data["data"])
        raise ValueError(f"Can't find pydantic model {data['__pydantic__']}")

    elif isinstance(data, dict) and "__django__" in data:
        app_name = data["__django__"]["app_name"]
        model_name = data["__django__"]["model_name"]
        pk = data["pk"]
        model_class = apps.get_model(app_name, model_name)
        return model_class.objects.get(pk=pk)

    elif isinstance(data, (list, tuple)):
        return [_deserialize(item, pydantic_models) for item in data]

    elif isinstance(data, dict):
        return {
            key: _deserialize(value, pydantic_models) for key, value in data.items()
        }

    return data



import logging
from prometheus_client import Counter

from promy.utils.monitoring import REGISTRY, METRICS_NAMESPACE

_LOG_LEVEL_TOTAL = Counter(
    "django_log_level_count",
    "Counts of logs by level",
    ["level"],
    registry=REGISTRY,
    namespace=METRICS_NAMESPACE,
)


class PrometheusLogHandler(logging.Handler):
    def emit(self, record):
        _LOG_LEVEL_TOTAL.labels(level=record.levelname.lower()).inc()



import logging
from typing import Any

from django.utils import timezone
from django_celery_beat.models import PeriodicTask, IntervalSchedule

from promy.utils.django import get_environment
from promy.utils.time import parse_duration

logger = logging.getLogger(__name__)


def set_periodic_tasks(data: dict | list, disable_rest: bool = True) -> None:
    if isinstance(data, list):
        tasks = data
    else:
        env = get_environment()
        if env is None:
            logger.error("Environment not set. Please set settings.ENV.environment")
            return
        if env not in data:
            return
        tasks = data[env]

    task_names = {"celery.backend_cleanup"}
    for task in tasks:
        if "name" not in task:
            logger.error("Periodic task is missing 'name' field")
            continue
        if task["name"] in task_names:
            logger.error(f"Duplicate periodic task name: {task['name']}")
            continue
        task_names.add(task["name"])

        _save_periodic_task(task)

    if disable_rest:
        _disable_remaining_periodic_tasks(task_names)


class PromTasks:
    PUBLISH_EVERYTHING = "promy.tasks.publish_everything"
    PUBLISH_RECENTLY_UPDATED_OBJECTS = "promy.tasks.publish_recently_updated_objects"
    UPDATE_EVERYTHING = "promy.tasks.update_everything"
    UPDATE_SET = "promy.tasks.update_set"

    @classmethod
    def all(cls) -> list[str]:
        res = []
        for attr_name, attr_value in cls.__dict__.items():
            if not attr_name.startswith("__") and isinstance(attr_value, str):
                res.append(attr_value)
        return res

    @classmethod
    def as_dict(cls) -> dict[str, str]:
        res = {}
        for attr_name, attr_value in cls.__dict__.items():
            if not attr_name.startswith("__") and isinstance(attr_value, str):
                res[attr_name] = attr_value
        return res


_UNSUPPORTED_PERIODIC_TASK_FIELDS = ("crontab", "solar", "clocked")
_ALL_PERIODIC_TASK_FIELDS = {
    "name": None,
    "task": None,
    "interval": None,
    "crontab": None,
    "solar": None,
    "clocked": None,
    "args": "[]",
    "kwargs": "{}",
    "queue": None,
    "exchange": None,
    "routing_key": None,
    "headers": "{}",
    "priority": None,
    "expires": None,
    "expire_seconds": None,
    "one_off": False,
    "start_time": None,
    "enabled": True,
    "description": "",
}


def _save_periodic_task(task_dict: dict) -> None:
    try:
        task = PeriodicTask.objects.get(name=task_dict["name"])
        changed = False
    except PeriodicTask.DoesNotExist:
        task = PeriodicTask(name=task_dict["name"])
        changed = True

    defaults = _ALL_PERIODIC_TASK_FIELDS | {"start_time": timezone.now()}

    for key, value in task_dict.items():
        setattr(task, key, _get_periodic_task_value(key, value))

    if "enabled" not in task_dict:  # enable task unless explicitly disabled
        task.enabled = True

    for key, value in defaults.items():
        if getattr(task, key, None) is None:
            setattr(task, key, value)

    if not changed:
        original_task = PeriodicTask.objects.get(name=task.name)
        for key in _ALL_PERIODIC_TASK_FIELDS:
            if getattr(original_task, key, None) != getattr(task, key, None):
                changed = True
                break

    if changed:
        logger.info(f"Saving periodic task: {task.name}")
        task.save()


def _get_periodic_task_value(key: str, value: Any) -> Any:
    if key == "interval":
        return _interval_schedule(value)
    if key in _UNSUPPORTED_PERIODIC_TASK_FIELDS:
        raise NotImplementedError(
            f"PeriodicTask with {key} is not supported yet. "
            f"Contact promy owners or implement it yourself and send us a Pull Request ;)"
        )
    return value


def _disable_remaining_periodic_tasks(task_names: set | list) -> None:
    for task in list(
        PeriodicTask.objects.filter(enabled=True).exclude(name__in=task_names)
    ):
        logger.info(f"Disabling unknown periodic task: {task.name}")
        task.enabled = False
        task.save()


def _interval_schedule(duration: str) -> IntervalSchedule:
    amount, unit = parse_duration(duration)
    if unit == "milliseconds":
        unit = "microseconds"
        amount *= 1000

    return IntervalSchedule.objects.get_or_create(every=amount, period=unit)[0]



from rest_framework.pagination import PageNumberPagination


class DefaultPagination(PageNumberPagination):
    page_size = 10  # Default page size
    page_size_query_param = "page_size"
    page_query_param = "page"
    max_page_size = 20  # Maximum allowed page size



import logging
import uuid

from django.db import models

_logger = logging.getLogger(__name__)


class BaseModelManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().filter(archived_at=None)


class BaseModel(models.Model):
    key = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    created_at = models.DateTimeField(auto_now_add=True, db_index=True)
    updated_at = models.DateTimeField(auto_now=True, db_index=True)
    archived_at = models.DateTimeField(null=True, blank=True)

    objects = BaseModelManager()
    all_objects = models.Manager()

    class Meta:
        default_manager_name = "objects"
        abstract = True
        ordering = ["-created_at"]

        def __init_subclass__(cls, **kwargs):
            super().__init_subclass__(**kwargs)

            meta_options_to_merge = [
                "constraints",
                "ordering",
                "indexes",
                "unique_together",
            ]

            # Merge specific Meta options from BaseModel.Meta
            for option in meta_options_to_merge:
                base_value = getattr(BaseModel.Meta, option, None)
                if base_value is None:
                    continue

                child_value = getattr(cls, option, None)

                if child_value is not None:
                    if isinstance(base_value, (list, tuple)):
                        merged_value = list(child_value) + [
                            item for item in base_value if item not in child_value
                        ]
                        setattr(cls, option, merged_value)
                    else:
                        logging.warning(
                            f"Meta option {option} is not a list or tuple, skipping merge"
                        )
                else:
                    setattr(cls, option, base_value)

    def __str__(self) -> str:
        return f"{self.__class__.__name__} [{self.pk}]"

    def save(self, *args, **kwargs):
        self.full_clean()
        super().save(*args, **kwargs)



from time import perf_counter as time_now
from unittest.mock import MagicMock

import redis.lock
from django.conf import settings

from promy.utils.exception import NotLockedError
from promy.utils.monitoring import REDIS_COMMAND_DURATION

_redis_client = None


class RedisClientWithMonitoring(redis.Redis):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def run_command(self, command: str, *args, **kwargs):
        start = time_now()
        success = "false"
        try:
            ret = getattr(super(), command)(*args, **kwargs)
            success = "true"
            return ret
        finally:
            REDIS_COMMAND_DURATION.labels(command, success).observe(time_now() - start)

    def append(self, *args, **kwargs):
        return self.run_command("append", *args, **kwargs)

    def bitcount(self, *args, **kwargs):
        return self.run_command("bitcount", *args, **kwargs)

    def bitfield(self, *args, **kwargs):
        return self.run_command("bitfield", *args, **kwargs)

    def bitfield_ro(self, *args, **kwargs):
        return self.run_command("bitfield_ro", *args, **kwargs)

    def bitop(self, *args, **kwargs):
        return self.run_command("bitop", *args, **kwargs)

    def bitpos(self, *args, **kwargs):
        return self.run_command("bitpos", *args, **kwargs)

    def copy(self, *args, **kwargs):
        return self.run_command("copy", *args, **kwargs)

    def decrby(self, *args, **kwargs):
        return self.run_command("decrby", *args, **kwargs)

    def delete(self, *args, **kwargs):
        return self.run_command("delete", *args, **kwargs)

    def dump(self, *args, **kwargs):
        return self.run_command("dump", *args, **kwargs)

    def exists(self, *args, **kwargs):
        return self.run_command("exists", *args, **kwargs)

    def expire(self, *args, **kwargs):
        return self.run_command("expire", *args, **kwargs)

    def expireat(self, *args, **kwargs):
        return self.run_command("expireat", *args, **kwargs)

    def expiretime(self, *args, **kwargs):
        return self.run_command("expiretime", *args, **kwargs)

    def get(self, *args, **kwargs):
        return self.run_command("get", *args, **kwargs)

    def getdel(self, *args, **kwargs):
        return self.run_command("getdel", *args, **kwargs)

    def getex(self, *args, **kwargs):
        return self.run_command("getex", *args, **kwargs)

    def getbit(self, *args, **kwargs):
        return self.run_command("getbit", *args, **kwargs)

    def getrange(self, *args, **kwargs):
        return self.run_command("getrange", *args, **kwargs)

    def getset(self, *args, **kwargs):
        return self.run_command("getset", *args, **kwargs)

    def incrby(self, *args, **kwargs):
        return self.run_command("incrby", *args, **kwargs)

    def incrbyfloat(self, *args, **kwargs):
        return self.run_command("incrbyfloat", *args, **kwargs)

    def keys(self, *args, **kwargs):
        return self.run_command("keys", *args, **kwargs)

    def lmove(self, *args, **kwargs):
        return self.run_command("lmove", *args, **kwargs)

    def blmove(self, *args, **kwargs):
        return self.run_command("blmove", *args, **kwargs)

    def mget(self, *args, **kwargs):
        return self.run_command("mget", *args, **kwargs)

    def mset(self, *args, **kwargs):
        return self.run_command("mset", *args, **kwargs)

    def msetnx(self, *args, **kwargs):
        return self.run_command("msetnx", *args, **kwargs)

    def move(self, *args, **kwargs):
        return self.run_command("move", *args, **kwargs)

    def persist(self, *args, **kwargs):
        return self.run_command("persist", *args, **kwargs)

    def pexpire(self, *args, **kwargs):
        return self.run_command("pexpire", *args, **kwargs)

    def pexpireat(self, *args, **kwargs):
        return self.run_command("pexpireat", *args, **kwargs)

    def pexpiretime(self, *args, **kwargs):
        return self.run_command("pexpiretime", *args, **kwargs)

    def psetex(self, *args, **kwargs):
        return self.run_command("psetex", *args, **kwargs)

    def pttl(self, *args, **kwargs):
        return self.run_command("pttl", *args, **kwargs)

    def hrandfield(self, *args, **kwargs):
        return self.run_command("hrandfield", *args, **kwargs)

    def randomkey(self, *args, **kwargs):
        return self.run_command("randomkey", *args, **kwargs)

    def rename(self, *args, **kwargs):
        return self.run_command("rename", *args, **kwargs)

    def renamenx(self, *args, **kwargs):
        return self.run_command("renamenx", *args, **kwargs)

    def restore(self, *args, **kwargs):
        return self.run_command("restore", *args, **kwargs)

    def set(self, *args, **kwargs):
        return self.run_command("set", *args, **kwargs)

    def setbit(self, *args, **kwargs):
        return self.run_command("setbit", *args, **kwargs)

    def setex(self, *args, **kwargs):
        return self.run_command("setex", *args, **kwargs)

    def setnx(self, *args, **kwargs):
        return self.run_command("setnx", *args, **kwargs)

    def setrange(self, *args, **kwargs):
        return self.run_command("setrange", *args, **kwargs)

    def stralgo(self, *args, **kwargs):
        return self.run_command("stralgo", *args, **kwargs)

    def strlen(self, *args, **kwargs):
        return self.run_command("strlen", *args, **kwargs)

    def substr(self, *args, **kwargs):
        return self.run_command("substr", *args, **kwargs)

    def touch(self, *args, **kwargs):
        return self.run_command("touch", *args, **kwargs)

    def ttl(self, *args, **kwargs):
        return self.run_command("ttl", *args, **kwargs)

    def type(self, *args, **kwargs):
        return self.run_command("type", *args, **kwargs)

    def watch(self, *args, **kwargs):
        return self.run_command("watch", *args, **kwargs)

    def unwatch(self, *args, **kwargs):
        return self.run_command("unwatch", *args, **kwargs)

    def unlink(self, *args, **kwargs):
        return self.run_command("unlink", *args, **kwargs)

    def lcs(self, *args, **kwargs):
        return self.run_command("lcs", *args, **kwargs)


def get_redis_client() -> RedisClientWithMonitoring:
    global _redis_client

    if _redis_client:
        return _redis_client

    redis_settings = settings.ENV.redis
    _redis_client = RedisClientWithMonitoring(
        host=redis_settings.host,
        port=redis_settings.port,
        db=redis_settings.db,
        username=redis_settings.user,
        password=redis_settings.password,
    )
    return _redis_client


def ensure_locked(lock: redis.lock.Lock) -> None:
    if lock.acquire(blocking=False):
        lock.release()
        raise NotLockedError()


def get_mock_redis():
    r = MagicMock()
    r.get = lambda *args, **kwargs: None
    return r



import re
from datetime import datetime, timedelta
from time import time
from typing import Tuple

try:
    from warnings import deprecated  # type: ignore
except ImportError:
    from deprecated import (
        deprecated,
    )  # TODO: remove this line after upgrading every project to python 3.13

from django.utils.timezone import make_aware

ISO_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S%z"
ISO_TIME_FORMAT_WITH_MS = "%Y-%m-%dT%H:%M:%S.%f%z"


def iso_serialize(dt: datetime, with_ms: bool = False) -> str:
    return dt.strftime(ISO_TIME_FORMAT_WITH_MS if with_ms else ISO_TIME_FORMAT)


def iso_deserialize(s: str) -> datetime:
    try:
        return datetime.strptime(s, ISO_TIME_FORMAT_WITH_MS)
    except ValueError:
        return datetime.strptime(s, ISO_TIME_FORMAT)


def expire_from_now(*, hours: int = 0, minutes: int = 0):
    return time() + hours * 3600 + minutes * 60


def get_datetime_from_timestamp_ms(ts: int, tz_aware: bool = True) -> datetime:
    dt = datetime.fromtimestamp(ts / 1000)
    return make_aware(dt) if tz_aware else dt


def get_timestamp_ms_from_datetime(dt: datetime) -> int:
    return int(dt.timestamp() * 1000)


def get_datetime_from_timestamp_s(ts: int, tz_aware: bool = True) -> datetime:
    return get_datetime_from_timestamp_ms(ts * 1000, tz_aware)


def get_timestamp_s_from_datetime(dt: datetime) -> int:
    return int(dt.timestamp())


@deprecated(
    "get_datetime_from_timestamp is deprecated because of ambiguous naming, use get_datetime_from_timestamp_ms or get_datetime_from_timestamp_s instead"
)
def get_datetime_from_timestamp(ts: int, tz_aware: bool = True) -> datetime:
    return get_datetime_from_timestamp_ms(ts, tz_aware)


@deprecated(
    "get_timestamp_from_datetime is deprecated because of ambiguous naming, use get_timestamp_ms_from_datetime or get_timestamp_s_from_datetime instead"
)
def get_timestamp_from_datetime(dt: datetime) -> int:
    return get_timestamp_ms_from_datetime(dt)


def parse_duration(time_str: str) -> Tuple[int, str]:
    match = re.match(r"^\d+$", time_str)
    if match:
        return int(time_str), "seconds"

    time_str = time_str.lower()

    time_mapping = {
        "s": "seconds",
        "m": "minutes",
        "h": "hours",
        "d": "days",
        "µs": "microseconds",
        "ms": "milliseconds",
        "us": "microseconds",
    }

    match = re.match(r"^(\d+)(ms|us|µs|[smhd])$", time_str)
    if not match:
        raise ValueError("Invalid time string format")

    amount, unit = match.groups()

    amount = int(amount)

    return amount, time_mapping[unit]


def parse_timedelta(time_str: str) -> timedelta:
    amount, unit = parse_duration(time_str)
    return timedelta(**{unit: amount})



from functools import wraps
from typing import Callable

from django.conf import settings
from opentelemetry import trace
from opentelemetry.sdk.resources import Resource, SERVICE_NAME
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk.trace.export import (
    BatchSpanProcessor,
)

tracing_settings = getattr(settings.ENV, "tracing", None)

resource = Resource(
    attributes={
        SERVICE_NAME: getattr(tracing_settings, "service_name", None)
        or settings.ENV.service_name
    }
)
provider = TracerProvider(resource=resource)
if getattr(tracing_settings, "enabled", False):
    processor = BatchSpanProcessor(
        OTLPSpanExporter(endpoint=getattr(tracing_settings, "url"))
    )
    provider.add_span_processor(processor)

trace.set_tracer_provider(provider)
tracer = trace.get_tracer(__name__)


def with_trace(name: str) -> Callable:
    def func(f: Callable) -> Callable:
        @wraps(f)
        def wrapped(*args, **kwargs):
            with tracer.start_as_current_span(name) as span:
                span.set_attribute("args", str(args))
                span.set_attribute("kwargs", str(kwargs))
                ret = f(*args, **kwargs)
                span.set_attribute("return", str(ret))
                return ret

        return wrapped

    return func



from django.http import HttpRequest, HttpResponse, HttpResponseServerError

from .health_check import health_check as hc


def health_check(request: HttpRequest) -> HttpResponse:
    result = hc.check_all()
    if result:
        return HttpResponse("OK")
    return HttpResponseServerError("ERROR")



import os

from celery import Celery

# Set the default Django settings module for the 'celery' program.
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "config.settings")

app = Celery("promy-default")

# Using a string here means the worker doesn't have to serialize
# the configuration object to child processes.
# - namespace='CELERY' means all celery-related configuration keys
#   should have a `CELERY_` prefix.
app.config_from_object("django.conf:settings", namespace="CELERY")

# Load task modules from all registered Django apps.
app.autodiscover_tasks()


@app.task(bind=True, ignore_result=True)
def debug_task(self):
    print(f"Request: {self.request!r}")



from promy.station.registry import station_registry
from promy.utils.celery import app
from promy.utils.monitoring import observe


@app.task
@observe("publish_everything", const_labels={"observer_type": "celery_task"})
def publish_everything() -> None:
    for publisher in station_registry.publishers:
        publisher.publish_all()


@app.task
@observe(
    "publish_recently_updated_objects", const_labels={"observer_type": "celery_task"}
)
def publish_recently_updated_objects() -> None:
    for publisher in station_registry.publishers:
        publisher.publish_recently_updated()


@app.task
@observe("update_everything", const_labels={"observer_type": "celery_task"})
def update_everything() -> None:
    if station_registry.periodic_updaters is None:
        return
    station_registry.periodic_updaters.update("all")


@app.task
@observe("update_set", const_labels={"observer_type": "celery_task"})
def update_set(name: str) -> None:
    if station_registry.periodic_updaters is None:
        return
    station_registry.periodic_updaters.update(name)



import os
import tempfile

from promy.utils.sentry import SentryHandler


def setup_settings(settings: dict) -> None:
    _install_additional_app(settings)
    _configure_project_settings(settings)
    _configure_celery_settings(settings)
    _configure_prometheus_settings(settings)
    _configure_sentry(settings)


def _install_additional_app(settings: dict) -> None:
    additional_apps = [
        "django_prometheus",
        "django_celery_results",
        "django_celery_beat",
    ]

    promy_idx = settings["INSTALLED_APPS"].index("promy")

    for app in additional_apps:
        if app not in settings["INSTALLED_APPS"]:
            settings["INSTALLED_APPS"].insert(promy_idx + 1, app)


def _configure_project_settings(settings: dict) -> None:
    if "ENV" not in settings:
        return
    env = settings["ENV"]

    if "DEBUG" not in settings:
        settings["DEBUG"] = env.debug
    if "SECRET_KEY" not in settings:
        settings["SECRET_KEY"] = env.secret_key


def _configure_celery_settings(settings: dict) -> None:
    if "ENV" not in settings:
        return
    env = settings["ENV"]

    if hasattr(env, "celery") and env.celery.enabled:
        if "CELERY_TASK_DEFAULT_QUEUE" not in settings:
            settings["CELERY_TASK_DEFAULT_QUEUE"] = (
                env.celery.default_queue
                if env.celery.default_queue
                else f"{env.service_name}-celery"
            )

        if "CELERY_RESULT_BACKEND" not in settings:
            settings["CELERY_RESULT_BACKEND"] = "django-db"

        if "CELERY_BEAT_SCHEDULER" not in settings:
            settings[
                "CELERY_BEAT_SCHEDULER"
            ] = "django_celery_beat.schedulers:DatabaseScheduler"

        if settings.get("TESTING"):
            settings["CELERY_BROKER_URL"] = "memory://"
            settings["CELERY_ALWAYS_EAGER"] = True
            settings["CELERY_TASK_EAGER_PROPAGATES"] = True

        else:
            settings["CELERY_BROKER_URL"] = env.celery.broker_url
            settings["CELERY_BROKER_CONNECTION_RETRY_ON_STARTUP"] = True


def _get_prometheus_db_engin_wrapper(engin: str):
    engin_mapping = {
        "django.db.backends.mysql": "django_prometheus.db.backends.mysql",
        "django.db.backends.postgresql": "django_prometheus.db.backends.postgresql",
        "django.db.backends.sqlite3": "django_prometheus.db.backends.sqlite3",
        "django.contrib.gis.db.backends.postgis": "django_prometheus.db.backends"
        ".postgis",
    }
    if engin in engin_mapping:
        return engin_mapping[engin]

    return engin


def _configure_prometheus_settings(settings: dict) -> None:
    if "ENV" not in settings:
        return
    env = settings["ENV"]

    if hasattr(env, "prometheus") and env.prometheus.enabled:
        if (
            "django_prometheus.middleware.PrometheusBeforeMiddleware"
            not in settings["MIDDLEWARE"]
        ):
            settings["MIDDLEWARE"].insert(
                0, "django_prometheus.middleware.PrometheusBeforeMiddleware"
            )
        if (
            "django_prometheus.middleware.PrometheusAfterMiddleware"
            not in settings["MIDDLEWARE"]
        ):
            settings["MIDDLEWARE"].append(
                "django_prometheus.middleware.PrometheusAfterMiddleware"
            )

        for db in settings["DATABASES"].values():
            db["ENGINE"] = _get_prometheus_db_engin_wrapper(db["ENGINE"])

        coordination_dir = env.prometheus.multiproc_dir
        if not coordination_dir:
            coordination_dir = tempfile.gettempdir() + "/prometheus-multiproc-dir/"
        os.makedirs(coordination_dir, exist_ok=True)
        os.environ["PROMETHEUS_MULTIPROC_DIR"] = coordination_dir


def _configure_sentry(settings: dict) -> None:
    if "ENV" not in settings:
        return
    env = settings["ENV"]

    if hasattr(env, "sentry"):
        SentryHandler.setup_sentry(env.sentry)



from django.db.models.query_utils import Q
from rest_framework import status
from rest_framework.decorators import api_view
from rest_framework.response import Response

from promy.station.publish import Publisher
from promy.station.registry import station_registry


@api_view(["GET"])
def query_publisher(request):
    publisher_name = request.query_params.get("name")
    publisher = _get_publisher_by_name(publisher_name)
    if isinstance(publisher, Response):
        return publisher

    query = Q()
    for k, v in request.query_params.items():
        if k == "name":
            continue
        query &= Q(**{k: v})

    qs = publisher.queryset.filter(query)
    messages = [publisher.get_instance_message(instance) for instance in qs]
    return Response({"results": messages}, status=status.HTTP_200_OK)


def _get_publisher_by_name(publisher_name: str) -> Publisher | Response:
    if not publisher_name:
        return Response(
            {"error": "name is required"}, status=status.HTTP_400_BAD_REQUEST
        )

    publisher = None
    for p in station_registry.publishers:
        if p.name == publisher_name:
            publisher = p
            break

    if not publisher:
        return Response(
            {"error": "publisher not found"},
            status=status.HTTP_404_NOT_FOUND,
        )

    return publisher



import logging
from rest_framework.views import exception_handler
from rest_framework import status
from rest_framework.response import Response
from requests.exceptions import RequestException

logger = logging.getLogger(__name__)

RETURN_TO_USER = "__return_to_user"


class NotLockedError(Exception):
    pass


class UserError(Exception):
    def __init__(self, msg, status_code: int = status.HTTP_400_BAD_REQUEST, **kwargs):
        self.msg = msg
        self.status_code = status_code
        self.kwargs = kwargs


class UserErrorResponse(Response):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        user_error_response(self)


class BaseControllerException(Exception):
    status = status.HTTP_400_BAD_REQUEST
    message = "<Base Controller Exception message>"
    data: dict = {}
    return_to_user = False

    def __init__(self, **kwargs):
        self.data = kwargs


def prom_exception_handler(exc, context):
    if isinstance(exc, UserError):
        return user_error_response(
            Response({"message": exc.msg, **exc.kwargs}, status=exc.status_code)
        )

    if isinstance(exc, BaseControllerException):
        resp = Response({"message": exc.message, "data": exc.data}, status=exc.status)
        if exc.return_to_user:
            resp = user_error_response(resp)
        return resp

    if isinstance(exc, RequestException):
        try:
            data = exc.response.json()
            if data.get(RETURN_TO_USER):
                return Response(data, status=exc.response.status_code)
        except Exception as e:
            logger.error(f"Request exception parsing error: {e=}, {exc.response=}.")

        logger.exception(
            f"Request exception: {exc=}, url={exc.response.url}, "
            f"status={exc.response.status_code}, response={exc.response.text}"
        )
        return Response(status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    # Default behaviour
    return exception_handler(exc, context)


def user_error_response(response: Response) -> Response:
    data = response.data or {}
    if isinstance(data, str):
        data = {"message": data}

    data[RETURN_TO_USER] = True
    response.data = data

    if response.status_code // 100 not in (4, 5):
        response.status_code = status.HTTP_400_BAD_REQUEST
    return response



from datetime import timedelta
from logging import getLogger

from promy.utils.cache import cache_get_key
from promy.utils.redis import get_redis_client


logger = getLogger(__name__)


def rate_limit_func(period: int | timedelta | None = None):  # period is in seconds
    def decorator(func):
        def wrapper(*args, **kwargs):
            final_period = period
            if final_period is None:
                final_period = kwargs.pop("rate_limit_period")

            if final_period is None:
                logger.error("rate_limit_period is not set")

            r = get_redis_client()
            cache_key = (
                "_promy:rate_limit:"
                + func.__name__
                + ":"
                + cache_get_key(*args, **kwargs)
            )
            if r.get(cache_key):
                return
            result = func(*args, **kwargs)
            if final_period:
                r.set(cache_key, 1, ex=final_period)
            else:
                r.set(cache_key, 1, ex=60 * 60 * 24)
            return result

        return wrapper

    return decorator



import logging
from abc import ABC, abstractmethod
from typing import List

from django.conf import settings
from django.db import connections
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.utils import OperationalError

from promy.utils.redis import get_redis_client

logger = logging.getLogger(__name__)


class BaseCheck(ABC):
    @abstractmethod
    def check(self) -> bool:
        """
        :return: True, if check is successful. False, otherwise.
        """
        pass


class DBCheck(BaseCheck):
    def __init__(self):
        super(DBCheck, self).__init__()

    def check(self) -> bool:
        try:
            conn: BaseDatabaseWrapper = connections["default"]
            conn.ensure_connection()
            return True
        except OperationalError as oe:
            logger.error(f"DB Check error: {oe}")
            return False


class KafkaCheck(BaseCheck):
    def __init__(self):
        super(KafkaCheck, self).__init__()

    def check(self) -> bool:
        # TODO: implement
        return True


class RedisCheck(BaseCheck):
    def __init__(self):
        super(RedisCheck, self).__init__()

    def check(self) -> bool:
        if (
            hasattr(settings, "ENV")
            and hasattr(settings.ENV, "redis")
            and getattr(settings.ENV.redis, "enabled", True)
        ):
            try:
                get_redis_client().ping()
                return True
            except ConnectionError as ce:
                logger.error(f"Redis Check error: {ce}")
                return False
        else:
            return True


class HealthCheck:
    def __init__(self) -> None:
        self._checks: List[BaseCheck] = []

    def add(self, c: BaseCheck) -> None:
        self._checks.append(c)

    def check_all(self) -> bool:
        for c in self._checks:
            if not c.check():
                return False
        return True


health_check = HealthCheck()
health_check.add(DBCheck())
health_check.add(KafkaCheck())
health_check.add(RedisCheck())



from urllib.parse import urljoin

import requests
from time import perf_counter as time_now

from promy.utils.monitoring import HTTP_CLIENT_DURATION

DEFAULT_TIMEOUT = 60


class HTTPClient:
    def __init__(
        self,
        service_name: str,
        session: requests.Session | None = None,
        base_url: str | None = None,
    ):
        self.service_name = service_name
        self._client = session or requests.Session()
        self.base_url = base_url

    def request(self, method: str, url: str, _api_name: str = "unset", **kwargs):
        kwargs.setdefault("timeout", DEFAULT_TIMEOUT)
        start = time_now()
        success = "false"

        try:
            response = self._client.request(method, url, **kwargs)
            success = "true"
            return response
        finally:
            HTTP_CLIENT_DURATION.labels(
                self.service_name, _api_name, method, success
            ).observe(time_now() - start)

    def get(self, url: str, _api_name: str = "unset", **kwargs):
        return self.request("GET", self._get_url(url), _api_name=_api_name, **kwargs)

    def post(self, url: str, _api_name: str = "unset", **kwargs):
        return self.request("POST", self._get_url(url), _api_name=_api_name, **kwargs)

    def put(self, url: str, _api_name: str = "unset", **kwargs):
        return self.request("PUT", self._get_url(url), _api_name=_api_name, **kwargs)

    def patch(self, url: str, _api_name: str = "unset", **kwargs):
        return self.request("PATCH", self._get_url(url), _api_name=_api_name, **kwargs)

    def delete(self, url: str, _api_name: str = "unset", **kwargs):
        return self.request("DELETE", self._get_url(url), _api_name=_api_name, **kwargs)

    def __setattr__(self, key, value):
        if key in ["service_name", "_client"]:
            super().__setattr__(key, value)
        else:
            setattr(self._client, key, value)

    def __getattr__(self, item):
        return getattr(self._client, item)

    def _get_url(self, url: str) -> str:
        if not self.base_url:
            return url
        if url.startswith("http://") or url.startswith("https://"):
            return url
        return urljoin(self.base_url, url)



from decimal import Decimal, ROUND_HALF_EVEN
from typing import Iterable, Any

EPSILON4 = 1e-1
EPSILON5 = 1e-5
EPSILON6 = 1e-6
EPSILON7 = 1e-7
EPSILON8 = 1e-8
EPSILON9 = 1e-9
EPSILON10 = 1e-10

EPSILON = EPSILON6


def to_decimal(value: float | int | str | Decimal) -> Decimal:
    if isinstance(value, Decimal):
        return value
    return Decimal(str(value))


def is_zero(value: float | Decimal, eps: float | Decimal = EPSILON) -> bool:
    return safe_abs(to_decimal(value)) < to_decimal(eps)


def safe_sum(col: Iterable[float | Decimal]) -> float:
    return float(sum([to_decimal(x) for x in col], Decimal(0)))


def safe_prod(col: Iterable[float | Decimal]) -> float:
    d = Decimal("1")
    for item in col:
        d *= to_decimal(item)
    return float(d)


def safe_add(a: float | Decimal, b: float | Decimal) -> float:
    return float(to_decimal(a) + to_decimal(b))


def safe_sub(a: float | Decimal, b: float | Decimal) -> float:
    return float(to_decimal(a) - to_decimal(b))


def safe_mult(a: float | Decimal, b: float | Decimal) -> float:
    return float(to_decimal(a) * to_decimal(b))


def safe_div(a: float | Decimal, b: float | Decimal) -> float:
    return float(to_decimal(a) / to_decimal(b))


def safe_abs(a: float | Decimal) -> float:
    return float(abs(to_decimal(a)))


def safe_eq(
    a: float | Decimal, b: float | Decimal, eps: float | Decimal = EPSILON
) -> bool:
    return is_zero(safe_sub(a, b), eps=eps)


def safe_ne(
    a: float | Decimal, b: float | Decimal, eps: float | Decimal = EPSILON
) -> bool:
    return not safe_eq(a, b, eps=eps)


def safe_gt(
    a: float | Decimal, b: float | Decimal, eps: float | Decimal = EPSILON
) -> bool:
    return safe_sub(a, b) > to_decimal(eps)


def safe_lt(
    a: float | Decimal, b: float | Decimal, eps: float | Decimal = EPSILON
) -> bool:
    return safe_sub(a, b) < -to_decimal(eps)


def safe_gte(
    a: float | Decimal, b: float | Decimal, eps: float | Decimal = EPSILON
) -> bool:
    return safe_sub(a, b) >= -to_decimal(eps)


def safe_lte(
    a: float | Decimal, b: float | Decimal, eps: float | Decimal = EPSILON
) -> bool:
    return safe_sub(a, b) <= to_decimal(eps)


def safe_max(a: float | Decimal, b: float | Decimal) -> float:
    return float(max(to_decimal(a), to_decimal(b)))


def safe_min(a: float | Decimal, b: float | Decimal) -> float:
    return float(min(to_decimal(a), to_decimal(b)))


def format_decimal(
    v: float | int | str | Decimal,
    decimals: int,
    *,
    rounding: str = ROUND_HALF_EVEN,
    normalize: bool = True,
) -> str:
    f = to_decimal(v).quantize(Decimal(f"1e-{decimals}"), rounding=rounding)
    if normalize:
        f = f.normalize()
    return format(f, "f")


def smart_format_decimal(
    v: float | int | str | Decimal,
    decimals: int,
    **kwargs: Any,
) -> str:
    """
    Format a decimal number in a context-aware way.
    Formats small numbers like 0.0000123456 with `decimals` digits after the leading zeros.
    For example with decimals=3, this would be 0.0000123, not 0 (what 'format_decimal' would yield).
    """
    d = to_decimal(v)
    s = format(d, "f")
    if not s.startswith("0."):
        return format_decimal(d, decimals, **kwargs)

    fractional_part = s.split(".")[1]
    zero_count = len(fractional_part) - len(fractional_part.lstrip("0"))
    return format_decimal(d, zero_count + decimals, **kwargs)



from django.urls import path

from promy.internal.views import health_check
from promy.station.views import query_publisher

urlpatterns = [
    path(
        "health",
        health_check,
        name="health_check",
    ),
    path("station/publisher/query", query_publisher),
]



import sentry_sdk
from sentry_sdk.integrations.django import DjangoIntegration

from promy.settings.env import SentrySettings


class SentryHandler:
    @staticmethod
    def traces_sampler(sampling_context):
        # This deactivates sending performance metrics (aka transactions) to the Sentry server
        # This was done because we do not need them and our Sentry instance cannot tolerate such a high-load
        return 0

    @staticmethod
    def setup_sentry(config: SentrySettings):
        if config.enabled:
            sentry_sdk.init(
                dsn=config.dsn,
                integrations=[
                    DjangoIntegration(),
                ],
                environment=config.environment,
                send_default_pii=True,
                traces_sampler=SentryHandler.traces_sampler,
            )



from functools import wraps
from time import perf_counter as time_now
from typing import Callable

from django.conf import settings
from prometheus_client import Counter, Histogram, multiprocess
from prometheus_client.registry import CollectorRegistry

__all__ = [
    "REGISTRY",
    "DURATION_SECONDS_BUCKETS",
    "METRICS_NAMESPACE",
    "Observer",
    "observe",
    "REDIS_COMMAND_DURATION",
    "KAFKA_CONSUME_DURATION",
    "KAFKA_PRODUCE_DURATION",
    "HTTP_CLIENT_DURATION",
]

REGISTRY = CollectorRegistry()
multiprocess.MultiProcessCollector(REGISTRY)

METRICS_NAMESPACE = settings.ENV.prometheus.prefix

DURATION_SECONDS_BUCKETS = (
    0.05,
    0.1,
    0.5,
    0.75,
    1.0,
    2.5,
    5.0,
    7.5,
    10.0,
    15.0,
    20.0,
    30.0,
    60.0,
    120.0,
    float("inf"),
)

DEPENDENCY_DURATION_SECONDS_BUCKETS = (
    0.005,
    0.01,
    0.05,
    0.075,
    0.1,
    0.25,
    0.5,
    0.75,
    1.0,
    1.5,
    2.0,
    3.0,
    6.0,
    12.0,
    float("inf"),
)


class Observer:
    def __init__(self, name: str, const_labels: dict | None = None) -> None:
        self.name = name
        self.const_labels = const_labels if const_labels else {}
        self.start: float | None = None

    @property
    def _labels(self) -> dict:
        return _get_observer_labels(self.name, self.const_labels)

    def set_labels(self, const_labels: dict) -> None:
        self.const_labels = const_labels

    def add_labels(self, const_labels: dict) -> None:
        self.const_labels.update(const_labels)

    def __enter__(self):
        _OBSERVER_ATTEMPTS_TOTAL.labels(**self._labels).inc()
        self.start = time_now()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        assert self.start
        success = "true" if not exc_type else "false"
        labels = self._labels
        if not labels.get("success"):
            labels["success"] = success

        _OBSERVER_TOTAL.labels(**self._labels).inc()
        _OBSERVER_DURATION.labels(**self._labels).observe(time_now() - self.start)


def observe(name: str, const_labels: dict | None = None) -> Callable:
    labels = _get_observer_labels(name, const_labels)

    def func(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            _OBSERVER_ATTEMPTS_TOTAL.labels(**labels).inc()
            start = time_now()
            success = "false"
            try:
                res = f(*args, **kwargs)
                success = "true"
                return res
            finally:
                if not labels.get("success"):
                    labels["success"] = success
                _OBSERVER_TOTAL.labels(**labels).inc()
                _OBSERVER_DURATION.labels(**labels).observe(time_now() - start)

        return wrapped

    return func


_OBSERVER_NAME_LABEL = "observer_name"

_OBSERVER_CONST_LABEL_NAMES = [
    "observer_type",
    "success",
    "method",
    "status",
    "api_endpoint",
    "api_name",
]

if hasattr(settings, "OBSERVER_EXTRA_LABEL_NAMES"):
    _OBSERVER_CONST_LABEL_NAMES.extend(getattr(settings, "OBSERVER_EXTRA_LABEL_NAMES"))
    _OBSERVER_CONST_LABEL_NAMES = list(set(_OBSERVER_CONST_LABEL_NAMES))

_OBSERVER_ATTEMPTS_TOTAL = Counter(
    f"observer_attempts_count_total",
    f"Observer attempts total count",
    registry=REGISTRY,
    labelnames=[*_OBSERVER_CONST_LABEL_NAMES, _OBSERVER_NAME_LABEL],
)  # type: ignore

_OBSERVER_TOTAL = Counter(
    f"observer_count_total",
    f"Observer total count",
    registry=REGISTRY,
    labelnames=[*_OBSERVER_CONST_LABEL_NAMES, _OBSERVER_NAME_LABEL],
)  # type: ignore

_OBSERVER_DURATION = Histogram(
    f"observer_duration_seconds",
    f"Observer duration (seconds)",
    registry=REGISTRY,
    buckets=DURATION_SECONDS_BUCKETS,
    labelnames=[*_OBSERVER_CONST_LABEL_NAMES, _OBSERVER_NAME_LABEL],
)  # type: ignore

REDIS_COMMAND_DURATION = Histogram(
    f"redis_duration_seconds",
    f"Redis duration (seconds)",
    buckets=DEPENDENCY_DURATION_SECONDS_BUCKETS,
    registry=REGISTRY,
    labelnames=["command", "status"],
)

KAFKA_CONSUME_DURATION = Histogram(
    f"kafka_consume_duration_seconds",
    f"Kafka consume duration (seconds)",
    buckets=DEPENDENCY_DURATION_SECONDS_BUCKETS,
    registry=REGISTRY,
    labelnames=["topic", "status"],
)

KAFKA_PRODUCE_DURATION = Histogram(
    f"kafka_produce_duration_seconds",
    f"Kafka produce duration (seconds)",
    buckets=DEPENDENCY_DURATION_SECONDS_BUCKETS,
    registry=REGISTRY,
    labelnames=["topic", "status"],
)

HTTP_CLIENT_DURATION = Histogram(
    f"http_client_duration_seconds",
    f"HTTP client duration (seconds)",
    buckets=DEPENDENCY_DURATION_SECONDS_BUCKETS,
    registry=REGISTRY,
    labelnames=["service_name", "api_name", "method", "status"],
)


def _get_observer_labels(observer_name: str, const_labels: dict | None) -> dict:
    labels = {_OBSERVER_NAME_LABEL: observer_name}
    if const_labels:
        labels.update(const_labels)

    for k in _OBSERVER_CONST_LABEL_NAMES:
        labels.setdefault(k, "")
    return labels



import logging
from abc import abstractmethod
from typing import TypeVar, Generic, Self, Any, Callable

from pydantic import BaseModel, ValidationError

from promy.utils.http_requests import HTTPClient
from promy.utils.models import BaseModel as DjangoBaseModel

ModelType = TypeVar("ModelType", bound=DjangoBaseModel)

logger = logging.getLogger(__name__)


class SkipUpdate(Exception):
    ...


class Updater:
    @abstractmethod
    def update(self):
        ...


class ModelUpdater(BaseModel, Generic[ModelType], Updater):
    @abstractmethod
    def update(self) -> ModelType:
        ...


class UpdaterHandler:
    def __init__(
        self,
        default_object_name: str | None = None,
        client: HTTPClient | None = None,
    ) -> None:
        self.handlers: dict[
            str, tuple[type[Updater], tuple[Callable[[Any], None], ...]]
        ] = {}
        self.default_object_name = default_object_name

        self.client = client
        if self.client:
            assert self.client.base_url, "base_url is not provided to HTTPClient"

    def add(
        self,
        object_name: str,
        updater_class: type[Updater],
        *callbacks: Callable[[Any], None],
    ) -> Self:
        assert object_name not in self.handlers
        self.handlers[object_name] = updater_class, callbacks
        return self

    def pull(self, name: str, **kwargs) -> list[Any]:
        assert self.client, "client is not provided to UpdaterHandler"
        resp = self.client.get(
            "station/publisher/query",
            params={"name": name, **kwargs},
        )
        resp.raise_for_status()
        resp_body = resp.json()
        results = resp_body["results"]
        instances = []

        for item in results:
            instance = self(item, default_name=name)
            if instance:
                instances.append(instance)

        return instances

    def pull_one(self, name: str, **kwargs) -> Any | None:
        instances = self.pull(name, **kwargs)
        if not instances:
            return None

        assert len(instances) == 1
        return instances[0]

    def __call__(
        self, value, *, raise_exception: bool = False, default_name: str | None = None
    ) -> Any | None:
        if "object_name" in value:
            name = value["object_name"]
            data = value["body"]

        elif default_name or self.default_object_name:
            name = default_name or self.default_object_name
            data = value
        else:
            logger.info("unknown object received")
            return None

        if name not in self.handlers:
            logger.info("unknown object received", extra={"object_name": name})
            return None

        updater_class, callbacks = self.handlers[name]

        try:
            data["updater_handler"] = self
            updater = updater_class(**data)
        except ValidationError as e:
            if raise_exception:
                raise e
            logger.exception("invalid object received")
            return None

        try:
            instance = updater.update()
        except SkipUpdate:
            logger.info("update skipped")
            return None

        if instance is None:
            return None

        if isinstance(instance, DjangoBaseModel):
            logger.info(
                "object persisted",
                extra={
                    "model": instance.__class__.__name__,
                    "key": instance.key,
                    "created_at": instance.created_at,
                    "updated_at": instance.updated_at,
                },
            )
        else:
            logger.info("object persisted", extra={"object_name": name})

        for callback in callbacks:
            try:
                callback(instance)
            except Exception as e:
                if raise_exception:
                    raise e
                logger.exception("callback failed")

        return instance



from pydantic import StrictStr, BaseModel
from pydantic_settings import BaseSettings, SettingsConfigDict


class EnvSettings(BaseSettings):
    debug: bool
    secret_key: StrictStr
    service_name: StrictStr
    environment: str | None = None

    model_config = SettingsConfigDict(
        env_prefix="",
        env_nested_delimiter="__",
        env_file=".env",
        extra="allow",  # Ignore extra fields
    )


class SentrySettings(BaseModel):
    enabled: bool = False
    dsn: str = ""
    environment: str = "production"
    sample_rate: float | int = 1.0


class LogstashSettings(BaseModel):
    enabled: bool = False
    host: str = None
    port: int = None
    version: int = 1


class KafkaSettings(BaseModel):
    enabled: bool = False
    bootstrap_servers: StrictStr
    security_protocol: StrictStr
    sasl_mechanism: StrictStr
    username: StrictStr
    password: StrictStr
    group_id: StrictStr


class PostgresSettings(BaseModel):
    name: StrictStr
    user: StrictStr
    password: StrictStr
    host: StrictStr
    port: int


class RedisSettings(BaseModel):
    host: StrictStr
    port: int
    db: int = 0
    user: StrictStr | None = None
    password: StrictStr | None = None


class CelerySettings(BaseModel):
    enabled: bool = False
    broker_url: StrictStr
    default_queue: StrictStr = ""


class PrometheusSettings(BaseModel):
    enabled: bool = False
    prefix: StrictStr
    multiproc_dir: StrictStr = "/tmp/multiproc-tmp"


class TracingSettings(BaseModel):
    enabled: bool = False
    url: StrictStr = ""
    service_name: StrictStr = ""
